#pragma once

/*
MIT License
Copyright (c) 2024 - 2025 René Amthor (tobid7)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
 */

// This file is generated by lazyvec
#include <pd/core/common.hpp>
#include <pd/core/vec2.hpp>  // Extended

namespace PD {
template <typename T>
class vec4 {
 public:
  T x;
  T y;
  T z;
  T w;

  vec4() : x(0), y(0), z(0), w(0) {}
  template <typename T1>
  explicit vec4(T1 v) {
    x = (T)v;
    y = (T)v;
    z = (T)v;
    w = (T)v;
  }

  template <typename T1>
  explicit vec4(vec4<T1> v) {
    x = (T)v.x;
    y = (T)v.y;
    z = (T)v.z;
    w = (T)v.w;
  }

  /** Extended Constructor */
  template <typename T1>
  explicit vec4(vec2<T1> a, vec2<T1> b) {
    x = (T)a.x;
    y = (T)a.y;
    z = (T)b.x;
    w = (T)b.y;
  }

  vec4(T x, T y, T z, T w) : x(x), y(y), z(z), w(w) {}

  template <typename T1>
  vec4<T>& operator+=(T1 v) {
    x += (T)v;
    y += (T)v;
    z += (T)v;
    w += (T)v;
    return *this;
  }

  template <typename T1>
  vec4<T>& operator+=(const vec4<T1>& v) {
    x += (T)v.x;
    y += (T)v.y;
    z += (T)v.z;
    w += (T)v.w;
    return *this;
  }

  template <typename T1>
  vec4<T> operator+(T1 v) const {
    return vec4<T>(x + (T)v, y + (T)v, z + (T)v, w + (T)v);
  }

  template <typename T1>
  vec4<T> operator+(const vec4<T1>& v) const {
    return vec4<T>(x + (T)v.x, y + (T)v.y, z + (T)v.z, w + (T)v.w);
  }

  template <typename T1>
  vec4<T>& operator-=(T1 v) {
    x -= (T)v;
    y -= (T)v;
    z -= (T)v;
    w -= (T)v;
    return *this;
  }

  template <typename T1>
  vec4<T>& operator-=(const vec4<T1>& v) {
    x -= (T)v.x;
    y -= (T)v.y;
    z -= (T)v.z;
    w -= (T)v.w;
    return *this;
  }

  template <typename T1>
  vec4<T> operator-(T1 v) const {
    return vec4<T>(x - (T)v, y - (T)v, z - (T)v, w - (T)v);
  }

  template <typename T1>
  vec4<T> operator-(const vec4<T1>& v) const {
    return vec4<T>(x - (T)v.x, y - (T)v.y, z - (T)v.z, w - (T)v.w);
  }

  template <typename T1>
  vec4<T>& operator*=(T1 v) {
    x *= (T)v;
    y *= (T)v;
    z *= (T)v;
    w *= (T)v;
    return *this;
  }

  template <typename T1>
  vec4<T>& operator*=(const vec4<T1>& v) {
    x *= (T)v.x;
    y *= (T)v.y;
    z *= (T)v.z;
    w *= (T)v.w;
    return *this;
  }

  template <typename T1>
  vec4<T> operator*(T1 v) const {
    return vec4<T>(x * (T)v, y * (T)v, z * (T)v, w * (T)v);
  }

  template <typename T1>
  vec4<T> operator*(const vec4<T1>& v) const {
    return vec4<T>(x * (T)v.x, y * (T)v.y, z * (T)v.z, w * (T)v.w);
  }

  template <typename T1>
  vec4<T>& operator/=(T1 v) {
    x /= (T)v;
    y /= (T)v;
    z /= (T)v;
    w /= (T)v;
    return *this;
  }

  template <typename T1>
  vec4<T>& operator/=(const vec4<T1>& v) {
    x /= (T)v.x;
    y /= (T)v.y;
    z /= (T)v.z;
    w /= (T)v.w;
    return *this;
  }

  template <typename T1>
  vec4<T> operator/(T1 v) const {
    return vec4<T>(x / (T)v, y / (T)v, z / (T)v, w / (T)v);
  }

  template <typename T1>
  vec4<T> operator/(const vec4<T1>& v) const {
    return vec4<T>(x / (T)v.x, y / (T)v.y, z / (T)v.z, w / (T)v.w);
  }

  vec4 operator-() const { return vec4(-x, -y, -z, -w); }

  bool operator==(const vec4& v) const {
    return x == v.x && y == v.y && z == v.z && w == v.w;
  }
  bool operator!=(const vec4& v) const { return !(*this == v); }

  double Len() const { return std::sqrt(SqLen()); }
  double SqLen() const { return x * x + y * y + z * z + w * w; }

  void SwapXY() {
    T t = x;
    x = y;
    y = t;
  }
  void SwapXZ() {
    T t = x;
    x = z;
    z = t;
  }
  void SwapXW() {
    T t = x;
    x = w;
    w = t;
  }
  void SwapYZ() {
    T t = y;
    y = z;
    z = t;
  }
  void SwapYW() {
    T t = y;
    y = w;
    w = t;
  }
  void SwapZW() {
    T t = z;
    z = w;
    w = t;
  }
};
using fvec4 = vec4<float>;
using dvec4 = vec4<double>;
using ivec4 = vec4<int>;
}  // namespace PD
